#include <sys_defs.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>                      /* strerror() */
#include <errno.h>
#include <string.h>
#include <mymalloc.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <vstream.h>

#include <iostuff.h>
#include "smtpd.h"

#define IDENT_MSGSIZE 256 
#define IDENT_TIMEOUT 10

#define CHOMP(STR) { char *tmp; tmp = STR; while (*tmp) { \
             if (*tmp == '\n' || *tmp == '\r') *tmp = '\0'; tmp++ ; } }

char *smtpd_ident(struct sockaddr_in *peer_addr, struct sockaddr_in *smtpd_addr)
{
    int ident_sock;
    char ident_msg[IDENT_MSGSIZE + 1], *sp;
    char ident_user[IDENT_MSGSIZE + 1];
    struct sockaddr_in local_addr;
    struct sockaddr_in ident_addr;
    char *return_val;
    VSTREAM *ident_stream;

    memset(ident_msg, 0, IDENT_MSGSIZE + 1);
    memset(ident_user, 0, IDENT_MSGSIZE + 1);

    /*
     * Bind the local sockaddr to the same interface as smtpd before
     * connecting back to the auth port on the peer. This helps
     * with multihomed postfix servers. First, set up the address.
     */

    /* Local sockname */

    memset((char *) &local_addr, 0, sizeof(local_addr));
    local_addr.sin_family = AF_INET;
    memcpy((void *) &local_addr.sin_addr, (void *) &smtpd_addr->sin_addr, sizeof(local_addr.sin_addr));

    /* Remote sockname + port */

    memset((char *) &ident_addr, 0, sizeof(ident_addr));
    ident_addr.sin_family = AF_INET;
    memcpy((void *) &ident_addr.sin_addr, (void *) &peer_addr->sin_addr, sizeof(ident_addr.sin_addr));
    ident_addr.sin_port = htons(113);

    do {
        /* socket call */

        if ((ident_sock = socket(ident_addr.sin_family, SOCK_STREAM, 0)) < 0) {
            msg_warn("Can't allocate socket for ident lookup: %s", strerror(errno));
            break;
        }

        /* Now bind the local sock to the interface */

        if (bind(ident_sock, (struct sockaddr *)&local_addr, sizeof(local_addr)) < 0) {
            msg_warn("local bind of ident sock failed: %s", strerror(errno));
            break;
         }

        /* connect() back to the smtp client host on port 113 */

         if (connect(ident_sock, (struct sockaddr *) &ident_addr, sizeof(ident_addr )) < 0) {
            msg_warn( "ident connect to %s: %s", inet_ntoa(peer_addr->sin_addr), 
                   strerror(errno));
            break;
         }

        /* Ok, make this a vstream */

        ident_stream = vstream_fdopen(ident_sock, O_RDWR);
        ident_stream->timeout = IDENT_TIMEOUT;

        /* Print the ident message to the remote host */
    
        vstream_fprintf(ident_stream, "%d, %d\n", ntohs(peer_addr->sin_port), ntohs(smtpd_addr->sin_port));
        if (vstream_ftimeout(ident_stream)) {
            msg_warn( "ident write timed out to %s", inet_ntoa(peer_addr->sin_addr));
            break;
        }

        /* Read back the result */

        vstream_fread(ident_stream, ident_msg, IDENT_MSGSIZE);
        if (vstream_ftimeout(ident_stream)) {
            msg_warn( "ident read timed out to %s", inet_ntoa(peer_addr->sin_addr));
            break;
        }
    
        /*
         * Should I even bother with this?
         *
         * Even if so, don't worry about this failing, set the timeout low
         */

        ident_stream->timeout = 2;
        vstream_fwrite(ident_stream, "quit\n", strlen("quit\n"));

        if (strlen(ident_msg) == 0) {
            msg_warn( "Failed to get ident string from %s", inet_ntoa(peer_addr->sin_addr));
            break;
        }
    
        if ((sp = strrchr(ident_msg, ':')) == NULL) {
            msg_warn( "Invalid ident string from %s", inet_ntoa(peer_addr->sin_addr));
            break;
        }
        sp++;
        CHOMP(sp);
        while (*sp && (*sp == ' ' || *sp == '\t')) {
            sp++;
        }

        /* If we break before this line, we know we had some sort of bad error */

        strncpy(ident_user, sp, IDENT_MSGSIZE);
        msg_info( "Received ident string %s from %s", sp, inet_ntoa(peer_addr->sin_addr));
    
    } while (0);

    if (strlen(ident_user) == 0) {
        msg_warn( "Failed to get ident user for %s", inet_ntoa(peer_addr->sin_addr));
        return NULL;
    } 
    
    vstream_fclose(ident_stream);
    return_val = mystrdup(ident_user);
    return return_val;
}
